% -*- mode: SLang; mode: Fold -*-
%
% A simple Hex editor mode for Jed
% Sun Sep 23 21:56:30 2001, by Dino Leonardo Sangoi.
%
% Usage:
% Open a file, call hexedit_mode. Edit it. Call hexedit_commit(0) to simply
% update the source buffer, of hexedit_commit(1) to save to file.
%
% Version 0.57
%
% (It can't do a lot right now, but is useable for some fast hex hacking).
%
% BUGS:
% The conversions (in and out) are REALLY slow!
%
_debug_info=1;
_traceback=1;

require("keydefs");
require("basecnv");
require("wordsrch");

% Static variables %{{{
static variable _strHex = "";
static variable _strChr = "";
static variable _ct = @Char_Type[256];
static variable _nBlocks = -1;
static variable _nBytes = -1;
static variable _dataStart = -1;
static variable HexEditMode = 0;
static variable heMark = "\t\t\t";
static variable HexEdit = "HexEdit";
%}}}

% Buffer building %{{{

% Write to the right buffer
static define _ins(b, s)
{
        whatbuf();
        setbuf(b);
        insert(s);
        setbuf(());
}

% Add to buffer 'b', at pos 'c' the char 's'
static define _bb_add(b, c, s)
{
	variable n = c mod 4;

	if (n == 0) {		% Block Boundary
		if (c mod _nBytes == 0) {	% Line full
			if (c) {
				_ins(b, _strHex+"\t\t\t"+_strChr);
				_strHex = "\n"+basecnv_itoa(c, 16, 8, '0') + "\t";
			} else
				_strHex = basecnv_itoa(c, 16, 8, '0') + "\t";
			_strChr = "";
		}
		_strHex += "\t";
	}
	_strHex += "\t"+basecnv_itoa(s, 16, 2, '0');
	_strChr += char(_ct[s]);
        return c+1;
}

% Fill last line
static define _bb_fill(b, c)
{
	variable n;

	while (c mod _nBytes) {
		if (c mod 4 == 0)
			_strHex += "\t";
		_strHex += "\t\t\t";
		c++;
	}
	_ins(b, _strHex+"\t\t\t"+_strChr);
}

%}}}

% Static functions %{{{

% Build a string with 'n' expressed in dec/hex/oct. (with width 'w')
static define _build_triple_num(n, w)
{
	return string(n)+",0x"+basecnv_itoa(n, 16, w, '0')+",0"+basecnv_itoa(n, 8, w, '0');

}

static define _hexedit_set_status(b, n)
{
	set_status_line("HexEdit: %b at " + _build_triple_num(b, 0) +
			" : " + _build_triple_num(n, 2) + "='" + char(n) + "' ", 0);
}

% Goto byte position 'n', in the panel 'hm' (0 = Hex, 1 = Chars)
define _hexedit_goto(n, hm)
{
	variable l = n / _nBytes;
	variable c;

	n -= l * _nBytes;
	if (hm == 0)
		c = 12 + (n/4) * 13 + (n mod 4) * 3;
	else
		c = _dataStart + n;

	goto_line(l+1);
	goto_column(c);
}

% get current position and char value
static define _hexedit_get_byte_info()
{
	variable l, x, b, s;

	l = what_line() - 1;
	x = what_column();
	if (HexEditMode) {
		x -= _dataStart;
		b = l * _nBytes + x;
		push_spot();
		_hexedit_goto(b, 0);
		s = get_word();
		pop_spot();
	} else {
		x -= 12;
		b = l * _nBytes + (x / 13) * 4 + (x mod 13) / 3;
		s = get_word();
	}

	return (b, basecnv_atoi(s, 16));
}

static define _hexedit_update()
{
	% Let the stack do the magic thing
	_hexedit_get_byte_info();
	_hexedit_set_status();
}

define _hexedit_bol()
{
	if (HexEditMode == 0)
		goto_column(12);
	else
		goto_column(_dataStart);
}

static define _hexedit_right_check()
{
	if (looking_at(heMark) or eolp()) {
		call("previous_char_cmd");
		call("next_line_cmd");
		_hexedit_bol();
	} else
		skip_chars("\t");
}

static define _hexedit_set_mode(m)
{
	variable b;

	(b,) = _hexedit_get_byte_info();
	HexEditMode = m;
	use_keymap(HexEdit+string(HexEditMode));
	_hexedit_goto(b, HexEditMode);
}


%}}}

% Movement functions %{{{
%move only in the char_table, skipping the tabs

define hexedit_bol()
{
	_hexedit_bol();
	_hexedit_update();
}

define hexedit_eol()
{
	if (HexEditMode == 0) {
		goto_column(_dataStart - 3);
		bskip_chars("\t");
	} else
		eol();
	go_left_1();
	_hexedit_update();
}

define hexedit_eob ()
{
	eob();
	hexedit_eol();
	% Update already done by hexedit_eol
}
define hexedit_bob ()
{
	bob();
	hexedit_bol();
	% Update already done by hexedit_bol
}

define hexedit_right ()
{
	call("next_char_cmd");
	_hexedit_right_check();
	_hexedit_update();
}

define hexedit_left ()
{
	if (blooking_at(heMark)) {
		call("previous_line_cmd");
		hexedit_eol();
	} else {
		call("previous_char_cmd");
		% This is safe for both sides
		if (looking_at_char('\t')) {
			bskip_chars("\t");
			go_left_1();
		}
	}
	_hexedit_update();
}

define hexedit_move_careful(f)
{
	variable c = what_column();

	call(f);
	() = goto_column_best_try(c);
	if (looking_at_char('\t') or eolp()) {
		call("previous_line_cmd");
		goto_column(c); % be safe
		error("End Of Buffer.");
	}
	_hexedit_update();
}

define hexedit_exit()
{
	delbuf(whatbuf());
}

define hexedit_ins()
{
	variable b, n;

	set_readonly(0);
	if (HexEditMode == 0)
		LAST_CHAR = toupper(LAST_CHAR);
	call("self_insert_cmd");
	push_spot();
	call("previous_char_cmd");
	(b, n) = _hexedit_get_byte_info();
	if (HexEditMode == 0) {
		LAST_CHAR=_ct[n];
		_hexedit_goto(b, 1);
	} else {
		n = what_char();
		_hexedit_goto(b, 0);
		variable s = basecnv_itoa(n, 16, 2, '0');
		LAST_CHAR = s[0];
		call("self_insert_cmd");
		LAST_CHAR = s[1];
	}
	call("self_insert_cmd");
	pop_spot();
	_hexedit_right_check();
	set_readonly(1);
	_hexedit_update();
}

define hexedit_undo()
{
	set_readonly(0);
	call("undo");
	_hexedit_right_check();
	set_readonly(1);
	_hexedit_update();
}

define hexedit_toggle()
{
	_hexedit_set_mode(1 - HexEditMode);
	_hexedit_update();
}

% commit changes to original buffer
define hexedit_commit()
{
	variable done = 0;
	variable s = "";
	variable hbuf = whatbuf();

	if (buffer_modified() == 0)
		message("Nothing to do!");
	else {
		_hexedit_set_mode(0);
		hexedit_bob();
		set_buffer_modified_flag(0);
		ERROR_BLOCK {
			_clear_error();
			done = 1;
		}
		while (done == 0) {
			s += char(basecnv_atoi(get_word(), 16));
			call("next_char_cmd");
			call("next_char_cmd");
			_hexedit_right_check();
		}

		variable Orig = hbuf;

		Orig = Orig[[1:-6]];
		flush("buf = "+Orig+". len = "+string(strlen(s))); sleep(2);
		setbuf(Orig);
		erase_buffer();
		insert(s);
	}
	setbuf(hbuf);
	hexedit_exit();
}
%}}}

% Mode initialisation %{{{
static define _set_keys(start, end, func, kmap)
{
	variable s;

	_for (start, end, 1) {
		s = char(());
		undefinekey(s, kmap);
		definekey(func, s, kmap);
	}
}

static define create_kmap(kmap, ext)
{
	if (keymap_p(kmap))
		return;

	make_keymap(kmap);

	% Editing
	if (ext == 0) {
		_set_keys('0', '9', "hexedit_ins", kmap);
		_set_keys('A', 'F', "hexedit_ins", kmap);
		_set_keys('a', 'f', "hexedit_ins", kmap);
	} else {
		_set_keys( 32, 126, "hexedit_ins", kmap);
		_set_keys(169, 255, "hexedit_ins", kmap);
	}
	% Movements
	definekey("hexedit_move_careful(\"previous_line_cmd\")",	Key_Up   , kmap);
	definekey("hexedit_move_careful(\"next_line_cmd\")",	Key_Down , kmap);
	definekey("hexedit_move_careful(\"page_up\")",		Key_PgUp , kmap);
	definekey("hexedit_move_careful(\"page_down\")",	Key_PgDn , kmap);
	definekey("hexedit_right",	Key_Right, kmap);
	definekey("hexedit_left",	Key_Left , kmap);
	definekey("hexedit_bol",	Key_Home , kmap);
	definekey("hexedit_eol",	Key_End  , kmap);
	definekey("hexedit_exit",       Key_F10, kmap);
	definekey("hexedit_undo",       "^_", kmap);
	definekey("hexedit_undo",       "^-", kmap);
	definekey("hexedit_toggle",     "\t", kmap);
}

create_syntax_table(HexEdit);
create_kmap(HexEdit+"0", 0);
create_kmap(HexEdit+"1", 1);

#ifdef HAS_DFA_SYNTAX
static define setup_dfa_callback (name)
{
	% for debugging
	% dfa_enable_highlight_cache("hexedit.dfa", name);
	dfa_define_highlight_rule("^[0-9A-F][0-9A-F]+", "keyword", name);
	dfa_define_highlight_rule("[0-9A-F][0-9A-F]\t", "number", name);
	dfa_build_highlight_table(name);
}
dfa_set_init_callback (&setup_dfa_callback, HexEdit);
%%% DFA_CACHE_END %%%
#endif

%}}}

% Main Routines %{{{
define hexedit()
{
        variable srcBuf = whatbuf();
        variable hexBuf = "*"+srcBuf+"-hex*";
	variable i;

	% Calculate buffer size
	_nBlocks = (SCREEN_WIDTH - 12) / 17;
	_nBytes = 4 * _nBlocks;
	_dataStart = 10 + 13*_nBlocks + 3;

	% Create escape table
	_for (0, 31, 1) {
		i = ();
		_ct[i] = '.';
	}
	_for (32, 255, 1) {
		i = ();
		_ct[i] = i;
	}
	_ct[127] = '.';

        if (bufferp(hexBuf))
                delbuf(hexBuf);
        setbuf(hexBuf);
        TAB = 1;
        setbuf(srcBuf);
        bob();
	% Build Buffer
	i = 0;
	flush("Reading...");
        do {
                i = _bb_add(hexBuf, i, what_char());
        } while (right(1));
	_bb_fill(hexBuf, i);

        sw2buf(hexBuf);
	set_overwrite(1);
	set_buffer_undo(1);
	set_buffer_modified_flag(0);
}

define hexedit_mode()
{
	hexedit();
#ifdef HAS_DFA_SYNTAX
	enable_dfa_syntax_for_mode(HexEdit);
#endif
	set_mode(HexEdit, 0);
	use_syntax_table(HexEdit);
	use_keymap(HexEdit+string(HexEditMode));
	bob();
	hexedit_bob();
	run_mode_hooks("hexedit_mode_hook");
}
%%%%%%%
%}}}
