% -*- mode: SLang; mode: fold -*-
_debug_info = 1;
_traceback=1;

%implements("Listings");

require("keydefs");

% Variables
variable ListingsMode = "Listings";

% Update hook to move line mark.
static variable ListingsLine = NULL;
custom_variable("ListingsSelectColor", color_number("menu_selection"));
custom_variable("ListingsMarkColor", color_number("region"));

static define listings_update_hook()
{
        ListingsLine = create_line_mark(ListingsSelectColor);
}

% Marks Support %{{{
static define _get_mark_info(mark)
{
        variable b, l, c;

        push_spot();    % Why I can access .line and .column of a Mark?
        whatbuf();
        b = user_mark_buffer(mark);
        setbuf(b);
        goto_user_mark(mark);
        l = what_line();
        c = what_column();
        setbuf(());
        pop_spot();
        return (b, l, c);
}

static variable Marks = @struct { mark, next };
Marks.mark = NULL; %  Sentinel
Marks.next = NULL;

%%%  Debug ON
define __dump_mark(i, m)
{
        variable b, l, c;

        whatbuf();
        setbuf("*mdump*");
        (b, l, c) = _get_mark_info(m);
        vinsert("%02d: \"%s\" \t % 6d:% 3d\n", i, b, l, c);
        setbuf(());
}

define _dump_marks()
{
        variable m, b, l, c, i = 1;

        foreach (Marks) {
                m = ();
                if (m.mark != NULL) {
                        __dump_mark(i, m.mark);
                }
                i++;
        }
}
%%%% Debug OFF

static define _find_mark(n)
{
        variable m, l;

        foreach(Marks) {
                m = ();
                if (m.mark != NULL) {
                        (, l,) = _get_mark_info(m.mark);
                        if (l == n)
                                return m;
                }
        }
        return NULL;
}

% Add a mark just before 'oldtag' (if oldtag == NULL, put at list head)
public define listings_tag_add(oldtag, mark)
{
        variable t = @struct { mark, next };

        if (oldtag == NULL)
                oldtag = &Marks;

        t.mark = mark;
        t.next = @oldtag;
        @oldtag = t;
}

public define listings_tag_set(rtag, mark)
{
        if (rtag == NULL)
                rtag = &Marks;

        (@rtag).mark = mark;
}

% Remove the 't' tag.
public define listings_tag_remove(t)
{
        if (t.next != NULL) {
                t.mark = t.next.mark;
                t.next = t.next.next;
        }
}

public define listings_tag_is_empty()
{
        return (Marks.mark == NULL);
}

public define listings_tag_count()
{
        variable m, i = 0;

        foreach (Marks) {
                m = ();
                if (m.mark)
                        i++;
        }
        return i;
}

public define listings_tag_clean()
{
        variable m;

        foreach (Marks) {
                m = ();
                if (m.mark == NULL) {
                        if (m.next != NULL) {
                                m.mark = m.next.mark;
                                m.next = m.next.next;
                        }
                }
        }
}


public define listings_tag_goto(t)
{
        if (t != NULL)
                if (t.mark != NULL) {
                        setbuf(user_mark_buffer(t.mark));
                        goto_user_mark(t.mark);
                }
}

public define listings_tag_untag_line()
{
        variable t;

        t = _find_mark(what_line());
        if (t == NULL) {
                % message("--- add tag ---"); sleep(2);
                listings_tag_add(NULL, create_line_mark(ListingsMarkColor));
        } else
                listings_tag_remove(t);
}

define listings_tag_foreach(func, param)
{
        variable m, ret = 0;
        % Assuming i'm already on the right buffer

        foreach (Marks) {
                m = ();
                if (m.mark != NULL) {
                        ret = @func(m, param);
                        if (ret != 0)
                                break;
                }
        }

        return ret;
}

%%% Example
static define _listings_remove(t, param)
{
        % This doesn't work well
        listings_tag_remove(t);
        return 0;
}

define listings_remove_tagged()
{
        return listings_tag_foreach(&_listings_remove);
}

public define listings_remove_line()
{
        set_readonly(0);
        bol();
        call("kill_line");
        set_buffer_modified_flag(0);
        set_readonly(1);
}

public define listings_remove_selected_lines()
{
        set_readonly(0);
        ERROR_BLOCK {
                set_buffer_modified_flag(0);
                set_readonly(1);
                _clear_error();
        }
        !if (listings_tag_is_empty()) {
                listings_remove_tagged();
        } else {
                if (markp()) {
                        narrow;
                        mark_buffer;
                        call("kill_region");
                        widen;
                } else
                        bol();
                % I should kill the void line left by kill_region...
                call("kill_line");
        }
        EXECUTE_ERROR_BLOCK;
}
%}}}

public define listings_follow()
{
        message("Listings doesn't know how to follow...");
}

create_syntax_table(ListingsMode);
!if (keymap_p (ListingsMode))
        make_keymap (ListingsMode);

definekey ("listings_tag_untag_line",              "t",     ListingsMode);  % Tag
definekey ("listings_tag_untag_line; go_down_1",   Key_Ins, ListingsMode);         % MC like
definekey ("listings_tag_untag_line; go_down_1",   Key_Shift_Down, ListingsMode);  % CUA style
definekey ("listings_tag_untag_line; go_up_1",     Key_Shift_Up,   ListingsMode);  % CUA style
definekey ("listings_follow",                "^M",    ListingsMode);  % Return
definekey ("delbuf(whatbuf)",               "q",     ListingsMode);
definekey ("delbuf(whatbuf)",               "\d155", ListingsMode);  % "meta-escape"
definekey ("listings_remove_selected_lines", "d",     ListingsMode);
definekey ("listings_remove_selected_lines", Key_Del, ListingsMode);  % CUA style

public define listings_mode ()
{
        set_buffer_modified_flag (0); % so delbuf doesnot ask whether to save first
        set_readonly(1);
        set_mode(ListingsMode, 0);
        use_keymap (ListingsMode);
        %%%%%% The real mode should do this
        %   use_syntax_table (ListingsMode);
        %   set_buffer_hook("mouse_2click", &listings_mouse_2click_hook);
        set_buffer_hook("update_hook", &listings_update_hook);
        run_mode_hooks("listings_mode_hook");
}

provide("listings");
