% number the buffers that have a file associated and bind
% Alt-# to goto_numbered_buffer

% comment these out once ready
%_traceback=1;
%_debug_info=1;

% Keybindings: Default is to bind Alt-1 to Alt-9 to goto_numbered_buffer.
% (With ALT_CHAR = 27) By default these are bound to digit_arg, make sure you
% bind another keyset to digit_arg.
% Use the following to change this. (Set to NULL if you don't want keybindings)
custom_variable("Numbuf_key_prefix", "\e");  % Alt/Escape/Meta

% Set this to 1 to have only Alt-0 bound to open_buffer_list()
custom_variable("Numbuf_show_list_when_failing", 1);

#ifnexists assoc_value_exists
% helper function:                (should be defined in a utility file)
% Find out if the associative array contains "value"
define assoc_value_exists (ass, value)
{
   foreach (ass) using ("values")
     if (() == value)
       return(1);
   return (0);
}
#endif

% now let's start:

variable Numbered_Buffer_List = Assoc_Type[String_Type];

% number the buffer if not done
define number_buffer ()
{
   % pop unwanted arguments when used as hook
   _pop_n (_NARGS);
   variable buf = whatbuf;
   % don't number hidden buffers and the ".jedrecent auxiliary buffer
   if (buf[0] == ' ' or buf == ".jedrecent")
     return;
   % find out if buffer is already numbered
   if(assoc_value_exists (Numbered_Buffer_List, buf))
     return;
   % find first free number
   variable i = 1;
   while (assoc_key_exists(Numbered_Buffer_List, string(i)))
     {
        if (bufferp(Numbered_Buffer_List[string(i)]))
          i++;
        else
          break;  % -> overwrite outdated entry
     }
   % add to list of numbered buffers and set status line
   Numbered_Buffer_List[string(i)] = buf;
   %   set_status_line("["+ string(i)+ "] %b  mode: %m %n  (%p)   %t ", 0);
   % This is not a perfect solution, but at least has some chances
   % to follow user preferences
   set_status_line("["+ string(i)+ "]" + Status_Line_String, 0);

}

% number every buffer that is associated to a file:
% either when opening a file (no arguments, no return value)
append_to_hook("_jed_find_file_after_hooks", &number_buffer);
% or when saving a up to now "unfiled" one to a file (one argument, no return value)
append_to_hook("_jed_save_buffer_after_hooks", &number_buffer);


% Build the menu of numbered buffers
static define change_buffer_callback (popup)
{
   loop (buffer_list ())
     {
        variable b = ();
        if (b[0] == ' ')
          continue;

        menu_append_item (popup, b, &sw2buf, b);
     }
}


public define numbered_buffers_menu_callback (popup)
{
   variable buf, key, cmd;
   variable menu;
   variable key_list = assoc_get_keys(Numbered_Buffer_List);

   foreach (key_list)
     {
        key = ();
        buf = Numbered_Buffer_List[key];
        if (bufferp(buf))
          menu_append_item (popup, "&"+key+" "+buf, "sw2buf", buf);
        else
          assoc_delete_key(Numbered_Buffer_List, key);
     }
   % append the unnumbered buffers
   loop (buffer_list ())
     {
        buf = ();
        if (buf[0] == ' ' or assoc_value_exists (Numbered_Buffer_List, buf))
          continue;
        if (buf[0] == '*')
          menu_append_item (popup, "*&"+buf[[1:]], &sw2buf, buf);
        else
          menu_append_item (popup, "&"+buf, &sw2buf, buf);
     }

}



% Change the callback of the Change Buffer menu entry
static define numbered_buffers_popup_hook (menubar)
{
   menu_set_select_popup_callback ("Global.&Buffers.&Change Buffer",
                                   &numbered_buffers_menu_callback);
}
append_to_hook ("load_popup_hooks", &numbered_buffers_popup_hook);

% Open Buffers List.
define open_buffer_list()
{
   ungetkey('c');
   ungetkey('b');
   call("select_menubar");
}

define goto_numbered_buffer(n)
{
   variable buf = "";
   if(assoc_key_exists(Numbered_Buffer_List, string(n)))
     buf = Numbered_Buffer_List[string(n)];
   if (bufferp(buf))
     sw2buf(buf);
   else if (Numbuf_show_list_when_failing)
     open_buffer_list();
   else
     message("Buffer "+string(n)+" doesn't exists.");
}

% Keybindings: Default is to bind Alt-1 to Alt-9 to goto_numbered_buffer
% See the custom variable Numbuf_key_prefix for changing this.
static define numbuf_bind_keys()
{
   variable i, digit;
   for(i=1; i<10; i++)
     {
        digit = string(i);
        setkey("goto_numbered_buffer("+ digit +")", Numbuf_key_prefix+digit);
     }
}

if (Numbuf_key_prefix != NULL)
{
   setkey("open_buffer_list", Numbuf_key_prefix + "0");
   numbuf_bind_keys();
}

provide("numbuf");

% test: message((assoc_get_values(Numbered_Buffer_List)))
