% number the buffers that have a file associated and bind
% Alt-# to goto_numbered_buffer

% comment these out once ready
%_traceback=1;
%_debug_info=1;

% Keybindings: Default is to bind Alt-1 to Alt-9 to goto_numbered_buffer.
% (With ALT_CHAR = 27) By default these are bound to digit_arg, make sure you
% bind another keyset to digit_arg.
% Use the following to change this. (Set to NULL if you don't want keybindings)
custom_variable("Numbuf_key_prefix", "\e");  % Alt/Escape/Meta

% Set this to 1 to have only Alt-0 bound to open_buffer_list()
custom_variable("Numbuf_show_list_when_failing", 1);

% helper function:                (should be defined in a utility file)
% Find out if the associative array contains "value"
define array_value_exists (arry, value)
{
   foreach (arry)
     if (() == value)
       return(1);
   return (0);
}

% now let's start:

variable Numbered_Buffer_List =String_Type[10];

% number the buffer if not done
define number_buffer ()
{
   % pop unwanted arguments when used as hook
   _pop_n (_NARGS);
   variable buf = whatbuf;
   % don't number hidden buffers and the ".jedrecent auxiliary buffer
   if (buf[0] == ' ' or buf == ".jedrecent")
     return;
   % find out if buffer is already numbered
   if(array_value_exists (Numbered_Buffer_List, buf))
     return;
   % find first free number
   variable i = 1;
   for (i=1; i < 10; i++)
     {
        if (Numbered_Buffer_List[i] == NULL)
          break;
        !if (bufferp(Numbered_Buffer_List[i]))
          break;  % -> overwrite outdated entry
     }
   % add to list of numbered buffers and set status line
   if (i < 10)
     {
        Numbered_Buffer_List[i] = buf;
        % This is not a perfect solution, but at least has some chances
        % to follow user preferences
        set_status_line("["+ string(i)+ "]" + Status_Line_String, 0);
     }
   else
     set_status_line("[*]" + Status_Line_String, 0);

}

% number every buffer that is associated to a file:
% either when opening a file (no arguments, no return value)
append_to_hook("_jed_find_file_after_hooks", &number_buffer);
% or when saving a up to now "unfiled" one to a file (one argument, no return value)
append_to_hook("_jed_save_buffer_after_hooks", &number_buffer);


% Build the menu of numbered buffers
public define numbered_buffers_menu_callback (popup)
{
   variable buf, key, cmd;
   variable menu;

   for (key = 1; key < 10; key++)
     {
        buf = Numbered_Buffer_List[key];
        if (buf == NULL)
          continue;
        if (bufferp(buf))
          menu_append_item (popup, "&"+string(key)+" "+buf, "sw2buf", buf);
        else
          Numbered_Buffer_List[key] = NULL;
     }
   % append the unnumbered buffers
   loop (buffer_list ())
     {
        buf = ();
        if (buf[0] == ' ' or array_value_exists (Numbered_Buffer_List, buf))
          continue;
        if (buf[0] == '*')
          menu_append_item (popup, "*&"+buf[[1:]], &sw2buf, buf);
        else
          menu_append_item (popup, "&"+buf, &sw2buf, buf);
     }

}

% Change the callback of the Change Buffer menu entry
static define numbered_buffers_popup_hook (menubar)
{
   menu_set_select_popup_callback ("Global.&Buffers.&Change Buffer",
                                   &numbered_buffers_menu_callback);
}
append_to_hook ("load_popup_hooks", &numbered_buffers_popup_hook);

% Open Buffers List.
define open_buffer_list()
{
   ungetkey('c');
   ungetkey('b');
   call("select_menubar");
}

define goto_numbered_buffer(n)
{
   variable buf = Numbered_Buffer_List[n];
   if (buf != NULL)
     if (bufferp(buf))
       {
          sw2buf(buf);
          return;
       }

   if (Numbuf_show_list_when_failing)
     open_buffer_list();
   else
     message("Buffer "+string(n)+" doesn't exists.");
}

% Keybindings: Default is to bind Alt-1 to Alt-9 to goto_numbered_buffer
% See the custom variable Numbuf_key_prefix for changing this.
static define numbuf_bind_keys()
{
   variable i, digit;
   for(i=1; i<10; i++)
     {
        digit = string(i);
        setkey("goto_numbered_buffer("+ digit +")", Numbuf_key_prefix+digit);
     }
}

if (Numbuf_key_prefix != NULL)
{
   setkey("open_buffer_list", Numbuf_key_prefix + "0");
   numbuf_bind_keys();
}

provide("numbuf");
