% c_bin_op.sl    -*- mode: SLang; mode: fold -*-
%
% Author: Dino Leonardo Sangoi
% Date: 10/11/2000
% Base on c_bin_op.sl from Manuel Arriaga
% Freely distributable
%
% Description:
%          A group of functions which tries to put spaces around binary operators.
%          while editing C code with jed.
%
% WARNING: These functions make mending code a true nightmare and, if you are lucky, they will
%          generate illegal statements whenever you try to do so; the other possibility is that
%          they will silently corrupt your program. That's why I've bound the "F2" key on my PC
%          to the unset_c_bin_op() function, which temporarily restores the normal keybindings.
%          c_bin_op can be reactivated by pressing "F2" again.
%          Using c_bin_op while working on code which wasn't written using it will probably cause
%          even more problems.
%          Last, but not least, please avoid using c_bin_op with files which are of importance to
%          you unless you've carefully read through this file and are certain that it won't
%          corrupt your program. I assume no liability for the consequences.
%
% Usage: Add the line
%
%        () = evalfile("c_bin_op");
%
%        to the C-mode hook in your .jedrc.
%
% Known bugs/problems/limitations:
%
% - oops, I don't know...
%
% I hope you enjoy it!
%
% Comments, additions, complaints and bug reports are welcome at
% g1001863@univ.trieste.it
% (original author: Manuel Arriaga <m.arriaga@ip.pt>)

%_debug_info = 1;
%_traceback  = 1;

implements("c_bin_op");

% Declare flag
variable C_BIN_OP;

% assoc. array for matching chars.
variable cbo = Assoc_Type[Struct_Type];

% Add an entry to cbo array, the arguments are :
% c   = character witch triggers the automatic spacing.
% str = String of chars witch inhibits spacing insertion.
%       (if == NULL never insert space, if == "" always)
% asp = 0 -> don't insert space after;
%       1 -> insert space after;
%       2 -> insert newline after;
static define mkcbo(c, str, asp)
{
	variable d = struct { str, asp };

	d.str = str;
	d.asp = asp;
	cbo[c] = d;
}

% Function definitions:
% Previous valid char...
static define pvc()
{
	variable c;

	push_mark();
	bskip_white();

	if (bolp())
	{
		pop_mark_1();
		return " ";
	}

	pop_mark_0();
	push_spot();
	go_left_1();
	c = char(what_char());
	flush( "<" + c + ">");
	pop_spot();
	return c;
}

static define _put_char(nc)
{
	variable c = pvc();
	variable nd;

	!if (assoc_key_exists(cbo, nc))
	{
		insert(nc);
		return;
	}

	nd =  cbo[nc];

	if ((nd.str != NULL) and (c != " "))
		!if (is_substr(nd.str, c))
			insert_char(' ');

	insert(nc);

	if (nd.asp == 2)
		newline();
	else if (nd.asp)
		insert_char(' ');
}

public define cbo_put_char()
{
	_put_char(char(LAST_CHAR));
	indent_line();
}

define unset_c_bin_op()
{
	variable keys = "=+-*!/|&,;?:><{";
	variable lkeys = strlen(keys);
	variable i;
	variable func;


	if (C_BIN_OP == 1)
		func = "self_insert_cmd";
	else
		func = "cbo_put_char";
	C_BIN_OP = 1 - C_BIN_OP;

	for (i = 0; i < lkeys; i++)
	{
		local_unsetkey(char(keys[i]));
		local_setkey(func, char(keys[i]));
	}
}

mkcbo("=", "=+-/*!", 1);
mkcbo("+", "+"     , 1);
mkcbo("-", "-"     , 1);
mkcbo("*", " *!%^&*([{" , 0);
mkcbo("!", ""      , 0);
mkcbo("/", ""      , 1);
mkcbo("|", "|"     , 1);
mkcbo("&", "&"     , 1);
mkcbo(",", NULL    , 1);
mkcbo(";", NULL    , 2);
mkcbo("?", ""      , 1);
mkcbo(":", ""      , 1);
mkcbo("<", "<"     , 1);
mkcbo(">", ">"     , 1);
mkcbo("{", ""      , 0);


% Bind "F2" to the function unset_c_bin_op()
% NO: Don't do this here, let the user choose the correct binding (if any).
%local_setkey("unset_c_bin_op", "\e[[B");

%Set flag
C_BIN_OP = 0;

% Activate c_bin_op:
unset_c_bin_op();
