%_traceback = 1;
%_debug_info = 1;


define c_make_doc_comment()
{
        insert("/* <DOC>\n * \n * </DOC>\n */");
        go_up(2);
}

%
%
%!%+
%\function{make_proto_ph}
%\synopsis{Make a Placeholder where get_protos() will insert prototypes}
%\usage{void make_proto_ph();}
%\description
%  This function puts two C comments at the cursor position.
%  These comments are used by \var{get_protos} to know where to
%  write prototypes for functions defined in the current file.
%\seealso{get_protos}
%!%-
define make_proto_ph()
{
        insert("\n/* *-* Prototypes ON *-* */\n/* *-* Prototypes OFF *-* */\n");
}

% Search for prototypes placeholder.
static define search_old_protos()
{
        variable s = NULL;
        bob();
        if (bol_fsearch("/* *-* Prototypes ON ")) {
                if (ffind(":")) {
                        go_right_1;
                        skip_white();
                        push_mark();
                        skip_chars("^ \t");
                        s = bufsubstr();
                }
                go_down_1;
                bol();
                push_mark();
                go_up_1;
                if (bol_fsearch("/* *-* Prototypes OFF *-* */")) {
                        call("kill_region");
                        return s;
                }
                pop_mark(1);
        }
        % Marker not found, as a fallback, go to the end of buffer.
        eob();
        insert("\n");
        go_left(1);
        return s;
}


% Put prototypes of functions defined in this file.
%!%+
%\function{get_protos}
%\synopsis{Put prototypes of functions defined in this file.}
%\usage{void get_protos();}
%\description
%  This function searchs for C functions in current buffer, and
%  writes a prototype for each function. the prototypes are written
%  Between two special comments (as created by \var{make_proto_ph}, or
%  at the end of file.
%  TODO: search for a matching .h file, and put there all the non-static prototypes. [DONE]
%\seealso{make_proto_ph}
%!%-
autoload("c_bskip_over_comment", "cmode");

static define is_static(str)
{
        variable s = is_substr(str, "static");
        variable c, slen;

        if (s < 1)
                return 0;

        slen = strlen(str);
        if (s != 1) {
                c = str[s-2];
                if ((c != ' ') and (c != '\t') and (c != '\n'))
                        return 0;
        }
        if (s+5 <= slen) {
                c = str[s+5];
                if ((c != ' ') and (c != '\t') and (c != '\n'))
                        return 0;
        }
        return 1;
}

define get_h_file(s)
{
        variable hm = NULL;

        if (s != NULL) {
                if (file_status(s) != 1)
                        s = NULL;
        }
        if (s == NULL) {
                % Look for .h file.
                s = path_sans_extname(buffer_filename());
                if (strlen(s) > 0) {
                        s = s + ".h";
                        if (file_status(s) != 1)
                                s = NULL;
                }
        }
        if (s != NULL) {
                read_file(s);
                s = whatbuf();
                () = search_old_protos();
                hm = create_user_mark();
        }
        return (s, hm);
}

define get_protos()
{
        variable b, m;
        variable hb = NULL, hm;
        variable s;
        variable c;

%        define_word("a-zA-Z0-9_");
        b = whatbuf();
        hb = search_old_protos();
        m = create_user_mark();

        (hb, hm) = get_h_file(hb);

        setbuf(b);
        eob();
        while (bsearch_char('}')) {
                parse_to_point();
                dup();
                if (() == -2) {
                        % Uhm, c_bskip_over_comment() doesn't skips if in middle
                        % of a C style comment, I try to found the end of comment
                        if (ffind("*/")) {
                                go_right(2);
                                c_bskip_over_comment(1);
                        } else {
                                if (bolp()) {
                                        go_up_1;
                                        eol();
                                } else
                                        bol();
                        }
                }
                if (() != 0)
                        continue;
                call("goto_match");
                c_bskip_over_comment(1);
                !if (re_bsearch("[]\\)\\}\\;\\>\\\"]"))
                        break;
                if (parse_to_point() != 0)
                        continue;
                c = what_char();
                if (c != ')') {
                        if (c == '}')
                                % Make sure this char will be matched by next bsearch_char()
                                go_right(1);
                        else  if ((c == ']') or (c == ')'))
                                call("goto_match");
                        continue;
                }
                right(1); % leave on Stack
                push_mark();
                go_left();  % get from stack
                call("goto_match");
                bskip_chars(" \t\n");
                bskip_word_chars();
                bskip_chars("^\n");
                s = bufsubstr();
                push_spot();
                goto_user_mark(m);
                if ((hb != NULL) and not is_static(s)) {
                        setbuf(hb);
                        goto_user_mark(hm);
                }
                vinsert("%s;\n", s);
                setbuf(b);
                pop_spot();
        }
        goto_user_mark(m);
}

% write the "#ifndef ..." "#define ..." "#endif" lines around a .h file
define c_make_h_file()
{
        variable hname = strup(str_replace_all(whatbuf(), ".", "_"));
        % Uhm, maybe I should be a bit more clever while generating
        % the define name.

        push_spot();
        bob();
        vinsert("#ifndef _%s_\n", hname);
        vinsert("#define _%s_\n\n", hname);
        eob();
        insert("\n#endif\n");
        pop_spot();
}

% Set to 1 to use old functions
%#define OLD_FIND_DELIMITER_FUNCS 1
#ifdef OLD_FIND_DELIMITER_FUNCS
static define _find_delimiter_up()
{
        variable r = 0;

        push_mark();
        forever {
                if (re_fsearch("[][(){}]") == 0) {
                        r = 1;
                        break;
                }
                if (find_matching_delimiter(0) != 1) {
                        r = 2;
                        break;
                }
                if (looking_at_char('(') or
                    looking_at_char('[') or
                    looking_at_char('{'))
                        break;
                right(1);
        }
        pop_mark(r);
        return r;
}

static define _find_delimiter_down()
{
        variable r = 0;

        push_mark();
        right(1);
        forever {
                if (re_bsearch("[][(){}]") == 0) {
                        r = 1;
                        break;
                }
                if (find_matching_delimiter(0) != 1) {
                        r = 2;
                        break;
                }
                if (looking_at_char(')') or
                    looking_at_char(']') or
                    looking_at_char('}'))
                        break;
                left(1);
        }
        pop_mark(r);
        return r;
}

#else
%
% Alternative versions of _find_delimiter_up/_down.
%

static define _match_valid_paren(search_f, match)
{
        variable r;

        forever {
                r = @search_f(match);
                if (r == 0)
                        break;
                if (parse_to_point() == 0)
                        break;
        }
        return r;

}

static define _fsearch_char_skip(match)
{
        go_right_1;
        return fsearch_char(match);
}

static define _re_fsearch_skip(regexp)
{
        go_right_1;
        return re_fsearch(regexp);
}

static define _find_delimiter_up()
{
        variable r = 0;

        push_mark();
        forever {
                if (_match_valid_paren(&re_bsearch, "[][(){}]") == 0) {
                        r = 1;
                        break;
                }
                if (looking_at_char('(') or
                    looking_at_char('[') or
                    looking_at_char('{'))
                        break;
                if (find_matching_delimiter(0) != 1) {
                        r = 2;
                        break;
                }
        }
        pop_mark(r);
        return r;
}

static define _find_delimiter_down()
{
        variable r = 0;

        push_mark();
        forever {
                if (_match_valid_paren(&_re_fsearch_skip, "[][(){}]") == 0) {
                        r = 1;
                        break;
                }
                if (looking_at_char(')') or
                    looking_at_char(']') or
                    looking_at_char('}'))
                        break;
                if (find_matching_delimiter(0) != 1) {
                        r = 2;
                        break;
                }
        }
        pop_mark(r);
        return r;
}

#endif

% This is how I would write this in C. I don't know if is a good
% thing to do in slang.
static variable _find_delimiter_errors = [ "",
                                          "could not find match",
                                          "could not find delimiter"];
define find_delimiter(dir)
{
        variable r;
        if (dir > 0)
                r = _find_delimiter_up();
        else
                r = _find_delimiter_down();
        if (r)
                error(_find_delimiter_errors[r]);
}

static variable LastLine = "make";
define sau_compile()
{
        variable line;

        line = read_mini("Compile:", "", LastLine);
        whatbuf();
        if (bufferp("makefile"))
                setbuf("makefile");
        else if (bufferp("Makefile"))
                setbuf("Makefile");
        else
                setbuf("*scratch*");
        compile(line);
	pop2buf(Compile_Output_Buffer);
        pop2buf(());
        LastLine = line;
}

define find_preproc(what, dir)
{
   variable s;
   variable f;

   if (what > 0) % open
      s = "^[ \t]*#[ \t]*if[ \tdn]";
   else % close
      s = "^[ \t]*#[ \t]*endif";

   if (dir > 0) {
      f = &re_fsearch;
      eol();
   } else
      f = &re_bsearch;

   do {
        if (@f(s) == 0)
         error("preproc open not found");
   } while (parse_to_point() != 0);
}


define set_compiler(comp)
{
   compile_select_compiler(comp);
   % variable f;


   % if (is_defined ("compile_parse_error_function"))
   %    compile_parse_error_function = comp;

   % f = __get_reference ("compile_parse_regexp");
   % if (f != NULL)
   %    @f = Null_String;
   % f = __get_reference ("Compile_Output_Buffer");
   % if (f != NULL)
   %    @f = Null_String;
   % () = evalfile("compile.sl");
}

define new_top_of_function()
{
        variable r;

        push_mark();
        r = _find_delimiter_up();
        if (r)
                if (_match_valid_paren(&bsearch_char, '}'))
                        r = _find_delimiter_up();

        pop_mark(r);
        if (r)
                verror("top of function not found.");
        else {
                forever {
                        if (_find_delimiter_up())
                                break;
                }
        }
}

define new_end_of_function()
{
        variable r;

        push_mark();
        r = _find_delimiter_down();
        if (r)
                if (_match_valid_paren(&_fsearch_char_skip, '{'))
                        r = _find_delimiter_down();

        pop_mark(r);
        if (r)
                verror("end of function not found.");
        else {
                forever {
                        if (_find_delimiter_down())
                                break;
                }
        }
}

define get_function_name()
{
        variable fn;

        push_mark();
        new_top_of_function();
        !if (bsearch_char(')'))
                break;
        right(1); % leave on Stack
        push_mark();
        go_left();  % get from stack
        call("goto_match");
        bskip_chars(" \t\n");
        bol();
        skip_chars(" \t");
        fn = bufsubstr();
        pop_mark(1);
        fn = str_delete_chars(fn, "\t\n");
        fn = str_replace_all(fn, "  ", " ");
        message(fn);
}

define create_h_file()
{
        variable b = whatbuf();
        variable s = path_sans_extname(buffer_filename());

        s = s + ".h";
        if (file_status(s) == 0) {
                read_file(s);
                % setbuf(s);
                make_proto_ph();
                c_make_h_file();
                write_buffer(s);
                setbuf(b);
                get_protos();
        }
}

% NYI

% This should make a list of fields in a structure, with the size and position
% of every field.
#iffalse
define get_field_list()
{
        if (find_matching_delimiter("}") != 1)
                return;
        push_mark();

        if (find_matching_delimiter(0) != 1)
                return;

        narrow();
        define_word("a-zA-Z0-9_");
        % to be continued....
        while (bol_bsearch_char('{')) {
                !if (bsearch_char(')'))
                        break;
                call("goto_match");
                bskip_chars(" \t\n");
                push_mark();
                call ("set_mark_cmd");
                bskip_word_chars();
                copy_region(saubuf);
                whatbuf();
                setbuf(saubuf);
                insert_char('\n');
                setbuf(());
        }
        call("kill_buffer");

}
#endif

define kill_function()
{
        push_mark();
        re_bsearch("[};]");
        go_down(1);
        exchange_point_and_mark();
        fsearch_char('{');
        call("goto_match");
        go_down(1);
%               ()=read_mini("wait", "", "");
        call("kill_region");
}

define c_reorder()
{
        variable mp;   % Mark for prototypes
        variable mi;   % Mark for Insert point;
        variable s;

        bob();
        !if (bol_fsearch("/* *-* Prototypes ON *-* */"))
                error("c_reoder: Can't find Prototypes mark.");
        go_down(1);
        mp = create_user_mark();
        !if (bol_fsearch("/* *-* Prototypes OFF *-* */"))
                error("c_reoder: Can't find Prototypes mark.");
        go_down(1);
        insert("\n");
        mi = create_user_mark();
        insert("\n");
        goto_user_mark(mp);
        go_right(1);
        forever {
                s = line_as_string();
                s = s[[0:-2]];
%               () = read_mini("s", s, "");
                !if (bol_fsearch(s))
                        break;
                kill_function();
                goto_user_mark(mi);
                go_up(1);
                call("yank");
%                insert("\n");
                goto_user_mark(mp);
                go_down(1);
                if (looking_at("/* *-* Prototypes OFF *-* */"))
                        break;
                move_user_mark(mp);
        }

%        () = read_mini("wait","","");
%        goto_user_mark(mp);
}

provide("cext");
