% cgram xref support for JED.
% by Dino Leonardo Sangoi <g1001863@univ.trieste.it>
%
% (What is cgram? is a program I'm slowly developing, when (if?) is at least
% a bit useable, I will release it.
%
%_debug_info=1;
%_traceback=1;

implements("Cgxref");

require("walk");
require("wordsrch");
require("bufmake");

custom_variable("Cgxref_Files", "cgxref");

static variable CgxrefBuffer = "*cgxref*";
static variable CgxrefBufferTime = 0;
static variable ChoiceBuffer = "*choice*";
static variable CgxrefKMap = "CgxrefKMap";
static variable CgxrefLine = NULL;

% Insert a line on Choice Buffer.
static define _ins_cb(str)
{
	whatbuf();
	setbuf(ChoiceBuffer);
	push_spot();
	bob;
	bol_fsearch(str); % On stack
	pop_spot();
	if (() == 0)
		insert(str+"\n");
	setbuf(());
}

% Update hook to move line mark.
static define cgxref_update_hook()
{
	CgxrefLine = create_line_mark(color_number("menu_selection"));
}

% Search given symbol on cgxref buffer, and build Choice buffer.
public define cgxref(symbol)
{
	if (symbol == NULL)
		return;
	if (strlen(symbol) == 0)
		return;
	walk_mark_current_position();

	bm_update_buffer(CgxrefBuffer, Cgxref_Files, &CgxrefBufferTime, NULL);

	symbol = symbol + "\t";
	if (bufferp(ChoiceBuffer)) {
		setbuf(ChoiceBuffer);
		set_buffer_modified_flag(0);
		delbuf(ChoiceBuffer);
	}
	setbuf(CgxrefBuffer);
	bob;
	if (bol_fsearch(symbol))
		do {
			_ins_cb(line_as_string());
			eol; % dont match again current line.
		} while (bol_fsearch(symbol));
	else {
		% not found
		walk_return_to_marked_position();
		message("cgxref: symbol not found: "+symbol);
		return;
	}

	setbuf(ChoiceBuffer);
	use_keymap(CgxrefKMap);
	set_buffer_hook("update_hook", &cgxref_update_hook);
	bob();
	walk_store_marked_position();
	walk_goto_current_position();
}

% cgxref current word.
public define cgxref_this()
{
	cgxref(get_word());
}

% Go to currently highlighted reference (bound to a key).
public define cgxref_goto()
{
	variable dl = strchop(line_as_string(), '\t', 0);

	walk_mark_current_position();
	if (read_file(dl[2])) {
		goto_line(integer(dl[3]));
		walk_store_marked_position();
		walk_goto_current_position();
	} else
		walk_return_to_marked_position();
}

!if (keymap_p(CgxrefKMap)) {
	make_keymap(CgxrefKMap);
	definekey("cgxref_goto", "
", CgxrefKMap);
}

provide("cgxref");
