%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% async Debugger Front-end for JED
%
% Based on ashell.sl
%
_traceback = 1;
_debug_info = 1;
% Matcher
() = evalfile("~/.jed/match.sl");

% PID
static variable ADBG_Id = -1;

static variable Debug_Lock = 0;
static variable Debug_Current_Command = NULL;
static variable Debug_Debugger = "dbxtra";

% Where
variable Debug_Conf_Dir = "~/.jed/";

define adbg_send_to_process(buf)
{
        % TODO: Prepare expect sequence Based on Command.
        send_process (ADBG_Id, buf + "\n");
}

define adbg_send_input ()
{
        variable buf;
        variable this_line, mark_line;
        variable m, ch, prompt;

        m = process_mark (ADBG_Id);

        this_line = what_line ();
        push_mark ();
        goto_user_mark (m);
        mark_line = what_line ();

        if (this_line >= mark_line) {
                pop_mark_0 ();
                push_mark_eob ();
                buf = bufsubstr ();
        } else {
                bskip_chars ("\t ");
                push_mark ();
                !if (bolp ()) {
                        go_left_1 ();
                        ch = what_char ();
                }
                bol ();
                prompt = bufsubstr ();
                pop_mark_1 ();
                bol ();
                if (looking_at (prompt)) {
                        go_right (strlen (prompt));
                }
                else if (ffind_char (ch)) {
                        go_right_1 ();
                }
                push_mark_eol ();
                buf = bufsubstr ();
                eob ();
                insert (buf);
        }
        newline ();
        move_user_mark (m);
        adbg_send_to_process(buf);
}

% Get Debugger Specific Info.
() = evalfile(strcat(Debug_Conf_Dir, Debug_Debugger));

define adbg_signal_handler (pid, flags, status)
{
        variable msg;

        eob ();
        msg = aprocess_stringify_status (pid, flags, status);
        vinsert ("\n\n----- %s ------\n\n", msg);
        ADBG_Id = -1;
}

define adbg_insert_output (pid, str)
{
        % TODO: check against expect sequence.
        () = match_matches(Debug_Current_Command.expect, str);
        Debug_Lock = 0;
        eob ();
        push_spot ();
        insert (str);
        pop_spot ();
        bol ();
        replace ("\r", Null_String);
        eob ();
        move_user_mark (process_mark (pid));
}

define adbg_put_string(str)
{
        eob();
        insert(str);
        insert("\n");
        adbg_send_to_process(str);
}

define adbg_put_command_internal(cmd)
{
        variable parsed_cmd;

        Debug_Current_Command = cmd;

        if (cmd.cmd != NULL) {
                if (cmd.getparm_func == NULL)
                        parsed_cmd = cmd.cmd;
                else {
                        @cmd.getparm_func();        % push Parameters
                        parsed_cmd = sprintf(cmd.cmd);
                }
                whatbuf(); % stack
                setbuf(debug_getbuf());
                adbg_put_string(parsed_cmd);
                setbuf(());
        }
}

define adbg_put_command(cmd)
{
        !if (Debug_Lock) {
                Debug_Lock = 1;
                dbg_clear_mark();
                adbg_put_command_internal(cmd);
        }
}


define debug_run()
{
        variable i, lenparm;

        dbg_ins("debug_run\n");
        lenparm = length(debug_params);
        for (i=0; i < lenparm; i++)
                debug_params[i];
        debug_program;
        lenparm;

        ADBG_Id = open_process ();
        set_process (ADBG_Id, "signal", "adbg_signal_handler");
        set_process (ADBG_Id, "output", "adbg_insert_output");

        adbg_put_command(dbgcmd_init);
}

define adbg()
{
        if (ADBG_Id != -1)
                error ("Only one Debug process is supported.");

        runhooks ("adbg_mode_hook");
        erase_buffer ();

        debug_run();

}

() = evalfile(strcat(Debug_Conf_Dir, "dbg_funs.sl"));

adbg();
