% Jed Debugger FrontEnd By saurus.
% Should Do these steps:
% init:
%    run the debugger With a config file.
%    (optional) Wait for Debugger Startup.
% 
% TODO:
%    - Create a Separate matcher module. - Done
%    - Add Pre and post matches. - Done.
%    - Support Multiple RegExps. - Done.
%    - Define a clean interface for Debugger Definitions.
%    - Generic Editor Buffer.
%    - Only One Buffer For Display. Display Editor Buffer.
%    - Native and Emulated Display Buffer.
%    - Breakpoint Editor Buffer.

% Matcher
() = evalfile("~/.jed/match.sl");
         
% Internal Type.
!if (is_defined("Debug_Cmd_Type")) {
	typedef struct {
		cmd,            % Command.
		getparm_func,   % Call to Get Params For function
		expect          % Match_Type Array
	} Debug_Cmd_Type;
}

% Internal Vars.
static variable dbg_pid = -1;
static variable Debug_Lock = 0;
static variable Debug_Curr_Cmd = NULL;

% These May Be Used By Debugger Specific Routines.
variable debug_buffer = "*trace*";
variable debug_display_buffer = "*display*";
variable debug_source_buffer = NULL;
variable debug_display_on = 0;
variable Debug_line_mark = 0;

% User Configurable Options 
variable Debug_Source_Lines = 15;             % Num Of Lines For The Source Window;
variable Debug_Conf_Dir = "/u/sdi/.jed/";
variable Debug_Debugger = "dbxtra";

define debug_output(pid, data) {
	
	setbuf(debug_buffer);
	insert(Sprintf("debug_output(%d) <%s>\n", pid, data, 2));
	() = match_matches(Debug_Curr_Cmd.expect, data);
%	if (match_matches(Debug_Curr_Cmd.expect, data)) {
%		setbuf(debug_buffer);
%		insert(sprintf("MATCH!!!!(%s)\n", Debug_Curr_Cmd.expect.regexp));
%	} else {
%		setbuf(debug_buffer);
%		insert(sprintf("Error Matching RE (%s)!\n", Debug_Curr_Cmd.expect.regexp));
%	}
	Debug_Lock = 0;
}

define debug_signal(pid, flags) 
{
	insert(Sprintf("debug_signal:parms: pid = <%d>, flags = <%d>\n", pid, flags, 2));
	Debug_Lock = 0;
}

% This Function Sets The Current Command Without Looking at The lock. Use only
% From Debugger Specific Functions (For Actions associated with keys, use
% debug_cmd() ).
define debug_cmd_internal(cmd) {
	variable parsed_cmd;

	setbuf(debug_buffer);
	if (cmd.cmd != NULL)
		insert(sprintf("debug_cmd: %s", cmd.cmd));
	Debug_Curr_Cmd = cmd;
	if (cmd.cmd != NULL) {
		if (cmd.getparm_func == NULL) 
			parsed_cmd = cmd.cmd;
		else
		{
			@cmd.getparm_func();        % push Parameters
			parsed_cmd = sprintf(cmd.cmd);
		}
		setbuf(debug_buffer);
		insert(sprintf("--- %s", parsed_cmd));
		send_process(dbg_pid, parsed_cmd);
	}
}

define debug_cmd(cmd) {
	variable parsed_cmd;

	!if (Debug_Lock) {
		Debug_Lock = 1;		
		Debug_line_mark = 0;
		debug_cmd_internal(cmd);
	}
}

% Get Debugger Specific Info.
() = evalfile(strcat(Debug_Conf_Dir, Debug_Debugger));

define debug_run()
{
	variable i, lenparm;

	setbuf(debug_buffer);
	insert("debug_run\n");
	lenparm = length(debug_params);
	for (i=0; i < lenparm; i++)
		debug_params[i];
	debug_program;
	lenparm;
	dbg_pid = open_process();
	set_process(dbg_pid, "output", "debug_output");
	set_process(dbg_pid, "signal", "debug_signal");
	debug_cmd(dbgcmd_init);
}

define sdefinekey(str, key, kmap)
{
	variable f;
        
        f = __get_reference (key);
        if (f != NULL)
                definekey(str, @f, kmap);
}

static variable DbgKMap = "DBGKMap";


define debug_init()
{
	make_keymap(DbgKMap);
	sdefinekey("debug_cmd(dbgcmd_step)",     "K_F8",  DbgKMap);
	sdefinekey("debug_cmd(dbgcmd_next)",     "K_F10", DbgKMap);
	sdefinekey("debug_cmd(dbgcmd_continue)", "K_F5",  DbgKMap);
	sdefinekey("debug_cmd(dbgcmd_goto)",     "K_F4",  DbgKMap);
	definekey("debug_cmd(dbgcmd_free)",     "",    DbgKMap);
	set_dbg_kmap(DbgKMap);
}

debug_init();

debug_run();
