%_debug_info = 1;
%_traceback = 1;
% create chords; each chord is a separate element of an associative array, and is referenced by the name
% of the chord.  The string assigned to the array element (which is parsed later in the script)
% specifies the strings of the chord.  The guitar strings are numbered 0 through 5 for the first band,
% 6 through 11 for the second, and so forth.  The set of six 0s and 1s denotes guitar strings that aren't
% played, and are shown flashing in the chart.
%  0  1  2  3  4  5
%  6  7  8  9 10 11
% 12 13 14 15 16 17
% 18 19 20 21 22 23
% 24 25 26 27 28 29
% 30 31 32 33 34 35


static variable chords = Assoc_Type[Struct_Type];
static variable ChordMarker = "%%% -*- ChordData -*-";
static variable ChordMarkerLen = strlen(ChordMarker);

define cmpint(a, b)
{
        return integer(a) - integer(b);
}

static define read_chords(filename)
{
        variable fp;
        variable FoundMark = 0;
        variable line, i;

        filename = expand_filename(filename);

        fp = fopen(filename, "r");
        if (fp == NULL)
                error("Error opening file \"" + filename + "\"");

        foreach (fp) using ("line") {
                line = ();
                if (FoundMark) {
                        % Read And Parse Data
                        % Format: ChordName;StringsToPlay;StringPos;StringPos;...
                        variable a;
                        variable d = struct { Play, Press };

                        a = strchop(line, ';', 0);

                        if (length(a) >= 3) {
                                for (i=1; i < length(a); i++)
                                        a[i] = strtrim(a[i]);
                                d.Play = a[1];
                                d.Press = a[[2:]];
                                chords[a[0]] = d;
                        }
                } else {
                        % Check For Marker

                        if (strncmp(line, ChordMarker, ChordMarkerLen) == 0)
                                FoundMark = 1;
                }

        }
        !if (FoundMark)
                error("Mark Not Found in file \"" + filename + "\"");
}

static define insert_band(start_pos, string)
{
        setbuf("*chords*");
        goto_column(start_pos);
        insert(string);
        if (down(1) != 1)
                insert("\n");
}

static define write_chord(chordname, start_pos)
{
        variable chord = chords[chordname];
        variable MaxBand = 5;
        variable PressNow;
        variable LineStart = "|||||||||||||||||||||||||||||||||||||||||||";
        variable Line = LineStart;
        variable Pos;
        variable P, I;
        variable i;

        insert_band(start_pos, chord.Play);
        insert_band(start_pos, "------");
        I = array_sort(chord.Press, &cmpint);
        P = chord.Press[I];
        foreach (P) {
                Pos = integer(());
                Line = strsub(Line, Pos + 1, '0');
        }

        for (i=0; i < 42; i = i + 6)
                insert_band(start_pos, Line[[i:i+5]]);
        insert_band(start_pos, "      ");
}

static define dump_array()
{
        variable k, c;
        variable pos, line;

        setbuf("*chords*");

        pos = 1;
        line = 1;
        foreach ( [ 'A', 'B', 'C', 'D', 'E', 'F', 'G' ] ) {
                c = ();
                foreach (chords) using ("keys") {
                        k = ();
                        if (k[0] == c) {
                                goto_line(line);
                                insert_band(pos, k);
                                insert_band(pos, "======");
                                write_chord(k, pos);
                                pos = pos + 8;
                        }
                }
                pos = 1;
                eob();
                line = what_line();
        }
}

read_chords("accordi.sl");
dump_array();

%
% function toggle(button) {
%        button.checked = !button.checked;
% }
%
% % remove the check from all radio buttons
% function resetGuitar() {
%        var Count;
%        clearTimeout(Timerid);
%        for (Count=1; Count < 42; Count++) {
%                document.guitar[Count-1].checked=false
%        }
% }
%
%
% % set the chord pattern
% function setGuitar() {
%        resetGuitar();
%        var Item, Ret, Count, Temp, Skip;
%        Item = document.guitar.chord.selectedIndex;
%        if (Item != -1) {
%                Text = document.guitar.chord.options[Item].text;
%                Frets = parser (chords[Text])
%                for (Count = 2; Count <= Frets[0]; Count++) {
%                        Temp = parseInt(Frets[Count]);
%                        document.guitar[Temp].checked=true;
%                }
%        }
% }
%
% % general function for parsing strings using a specified parse character; result is values in
% % separate elements of an array
% function parser (InString)  {
%        var Sep = ";", NumSeps=1, Count, Start, ParseMark, parse;
%        for (Count=1; Count < InString.length; Count++)  {
%                if (InString.charAt(Count)==Sep)
%                        NumSeps++;
%        }
%        parse = new Array ();
%        var Start=0, Count=1, ParseMark=0, LoopCtrl=1;
%        while (LoopCtrl==1)  {
%                ParseMark = InString.indexOf(Sep, ParseMark);
%                TestMark=ParseMark+0;
%                if ((TestMark==0) || (TestMark==-1)){
%                        parse[Count]= InString.substring (Start, InString.length);
%                        LoopCtrl=0;
%                        break;
%                }
%                parse[Count] = InString.substring (Start, ParseMark);
%                Start=ParseMark+1, ParseMark=Start, Count++;
%        }
%        parse[0]=Count;
%        return (parse);
% }
%

#iffalse
%%% -*- ChordData -*-
A;100000;1;5;14;15;16
Asus2;100000;1;4;5;14;15
Asus4;100000;1;4;14;15;22
A6;100000;1;14;15;16;17
A7;100000;1;3;5;14;16
Amaj7;100000;1;5;9;14;16
Am;100000;1;5;10;14;15
Am7;100000;1;14;3;10;5
Am7/G;010000;3;5;10;14;18
A13;100000;1;3;20;22;23
Bb;100000;7;20;21;22;11
Bbm;100000;7;20;21;16;11
Bbm7;100000;7;20;9;16;11
Bbmaj7;100000;7;20;15;22;11
B;100000;13;26;27;28;17
B7;100000;13;26;15;28;17
Bm;100000;13;26;27;22;17
Bm7;100000;13;26;15;22;17
Bsus4;100000;13;26;27;34;17
C;100000;19;14;3;10;5
C7;100000;19;14;21;10;5
Cm;100000;19;32;33;28;23
Cm7;100000;19;32;21;28;23
Cmaj7;100000;19;14;3;4;5
C/B;101000;13;3;10;5
C/G;000000;18;19;14;3;10;5
Cadd9;100000;19;14;3;22;5
CaddG;100000;19;14;3;10;23
C#;100000;25;38;39;40;29
C#m;100000;25;38;39;34;29
C#m7;100000;25;38;27;34;29
D;110000;2;15;17;22
D7;110000;2;10;15;17
Dm;110000;2;11;15;22
D9;100000;26;31;33;34;35
Dm7;110000;2;15;10;11
D/F#;000000;12;1;2;15;22;17
Dsus2;110000;2;15;22;5
Dsus4;110000;2;15;22;23
Eb;110000;32;21;28;23
Eb9;100000;20;25;27;28;29
E;000000;0;13;14;9;4;5
Em;000000;0;13;14;3;4;5
Em7;000000;0;13;14;3;22;5
E7;000000;0;13;2;9;4;5
E/G#;010000;24;14;27;4;5
Esus4;000000;0;13;14;15;4;5
E13;000000;0;2;5;9;13;16
F;000000;6;19;20;15;10;11
Fm;000000;6;19;20;9;10;11
Fm7;000000;6;19;8;9;10;11
Fmaj7;000000;6;1;20;15;10;5
F/G;000000;18;1;20;15;10;11
F#;000000;12;25;26;21;16;17
F#m;000000;12;25;26;15;16;17
F#m7;000000;12;25;14;15;16;17
F#7;000000;12;25;14;21;16;17
G;000000;2;3;4;13;18;23
Gm;000000;18;31;32;21;22;23
Gm7;000000;18;31;20;21;22;23
G6;010000;18;14;3;4;23
G7;000000;18;13;2;3;4;11
Gmaj7;010000;18;2;3;22;17
G9;000001;18;1;20;15;22
Gsus4;010000;18;2;3;10;23
G/B;100000;13;2;3;22;23
G/F#;010000;12;2;3;4;23
G#;000000;24;37;38;33;28;29
G#7;000000;24;37;26;33;28;29
G#m;000000;24;37;38;27;28;29
G#m7;000000;24;37;26;27;28;29
G#sus4;000000;24;37;38;39;28;29

#endif
