% ntte, New Tecnology Tags Engine.
% -*- mode: SLang; mode: folding -*-
%
_debug_info = 1;
_traceback = 1;

implements("Ntte");

() = evalfile("~/.jed/massoc.sl");

typedef struct { file, proto, n, type } TagInfo_Type;

% User-definable variables %{{{
custom_variable("Ntte_Add_Path_To_Tag", 0);
custom_variable("Ntte_Follow_Now", 1);
custom_variable("Ntte_Case_Sensitive", 1);
custom_variable("Ntte_Never_Ask", 1);
custom_variable("Ntte_Look_From_End", 1);
%}}}

static variable NtteOldTag = "";
static variable NtteOldPos = -1;
static variable NtteOldFile = "";

% "Low Level" Tag retrieving functions %{{{

% Get the tag under cursor, and ask user if configured to.
static define ntte_get_tag()
{
        variable tag = "0-9A-Z_a-z";
#ifdef VMS
        tag = strcat(tag, "$");
#endif
        push_spot();
        bskip_chars(tag);
        push_mark();
        skip_chars(tag);
        bufsubstr();   % Leave on stack
        pop_spot();

        if (Ntte_Follow_Now == 0)
                strtrim(read_mini("Find tag:", "", ())); % Leave on stack

        % Return value on stack
}

% Get info from a tag line.
static define ntte_parse_tag_line(line)
{
	variable tag = "";
        variable file = "";
	variable proto = "";
	variable n = -1;
        variable pos, len;

        if (string_match(line, "^\\([a-zA-Z_][a-zA-Z0-9_]*\\)\t\\([^\t]+\\)\t\\(/[^/]+/\\)", 1)) {
		%               +----------tag--------------+  +---file---+  +---proto---+
		% Pattern Tag, get prototype
                (pos, len) = string_match_nth(3);
		if (len > 2) {
			len = len - 2 - (line[pos+len-2] == '$');
			proto = str_replace_all(line[[pos+2:pos+len]], "\\/", "/");
			n = 0;
		}
        } else if (string_match(line, "^\\([a-zA-Z_][a-zA-Z0-9_]*\\)\t\\([^\t]+\\)\t\\([0-9]+\\)", 1)) {
		%                      +----------tag--------------+  +---file---+  +----num---+
                % Numeric Tag...
		% "quoted" tag , TAB , ( NOT_TAB+ ) , TAB + ( NUMBER+ )
                (pos, len) = string_match_nth(2);
                n = integer(line[[pos:pos+len-1]]);
                proto = "";
        }

        if (n >= 0) {
		% Get Tag
		(, len) = string_match_nth(1);
		tag = line[[:len-1]];
		% Get File
                (pos, len) = string_match_nth(2);
                file = line[[pos:pos+len-1]];
        }

        return (tag, file, proto, n);
}

% Get Tag Type set by Exhuberant ctags.
static define ntte_get_type(line)
{
	% type[0]:
	%              ' '  unknown
	%               c   classes
	%               d   macro definitions (and #undef names)
	%               e   enumerators
	%               f   function definitions
	%               g   enumeration names
	%               m   class, struct, or union members
	%               n   namespaces
	%               p   function prototypes and declarations [off]
	%               s   structure names
	%               t   typedefs
	%               u   union names
	%               v   variable definitions
	%               x   extern and forward variable declarations [off]
	%
	% type[1]:
	%               s   static
	%
        variable type = "  ";
        variable pos = is_substr(line, "/;\"");

        if (pos) {
		char(line[pos+3]);
		if (is_substr(line[[pos+5:]], "file:"))
			"s";
		else
			" ";
		() + ();
		type = ();
        }
        return type;
}
%}}}

% Tags Files Handling. %{{{

static variable TagsArray = NULL;
static variable TagsArrayTime = 0;

% Check if any of files has been modified.
static define ntte_files_changed(Time)
{
        variable s;

        foreach (strchop(Tag_Files, ',', 0)) {
                s = stat_file(());
                if (s != NULL)
                        if (s.st_mtime > Time)
                                return 1;
        }
        return 0;
}

% read a tag file and fill Tags Array.
static define ntte_read_file(fname)
{
	variable fp;
	variable tag;
	variable data = @TagInfo_Type;
	variable fdir = "";
	variable line;

        fname = expand_filename(fname);

	fp = fopen(fname, "r");
	if (fp == NULL)
		return ; % silently

        flush("ntte: rebuilding tags array from file ("+fname+")...");

	if (Ntte_Add_Path_To_Tag)
		(fdir, ) = parse_filename(fname);

	foreach (fp) using ("line") {
		line = ();
		(tag, data.file, data.proto, data.n) = ntte_parse_tag_line(line);

		data.file = fdir + data.file;
		data.type = ntte_get_type(line);
		if (Ntte_Case_Sensitive == 0)
			tag = strup(tag);

		if (data.n >= 0)
			massoc_insert(TagsArray, tag, @data);
	}
	() = fclose (fp);
}

% Check if tag buffer needs update, and do it.
static define ntte_update_array()
{
        variable redo_tags = 1;

        if (TagsArray != NULL) {
                if (TagsArrayTime != 0)
                        redo_tags = ntte_files_changed(TagsArrayTime);
        }

        if (redo_tags) {
                flush("ntte: rebuilding tags array...");
		TagsArray = massoc_create();
                foreach (strchop(Tag_Files, ',', 0))
                        ntte_read_file(());
                TagsArrayTime = _time();
                message("ntte: rebuilding tags array... Done.");
        }
}
%}}}

% Test Function %{{{
define testit()
{
	variable c1 = 0;
	variable c2 = 0;
	variable ct = 0;
	variable a, k, data;

	setbuf("*scratch*");

	flush("Reading...");
	ntte_update_array();
	flush("Reading... Done. Dumping....");

	foreach (TagsArray) using ("keys", "values") {
		c2 = 0;
		(k, a) = ();
		insert(k + " = \n");
		foreach (a.head) {
			data = ();
			if (data.data.n > 0)
				insert("	[ " + data.data.file+":"+ data.data.n + "]\n");
			else
				insert("	[ " + data.data.file+" : "+ data.data.proto + "]\n");
			ct++;
			c2++;
		}
		insert(" -- ["+string(c2)+"]\n");
		c1++;
	}
	insert("c1="+string(c1)+"\n");
	insert("ct="+string(ct)+"\n");

	_print_stack();

	flush("Dumping... Done.");
}

testit();
%}}}

% "High level" tags functions %{{{
static define ntte_calculate_score(current_tag)
{
        variable only_file;
	variable type;
	variable score = 0;

	return score;

	if (current_tag.type[0] != 'p')
		score += 1000;
}

% ctags format:
%  function-name\tfilename\t/^function-prototype/
%  typedef-name\tfilename\tline-number
% This Function Returns (filename, expresion, linenum)
% if linenum < 0 Tag not found.
% if linenum > 0 Tag Found as name + line
% if linumun == 0 Tag found as expresion.
static define ntte_find(start_file, tag)
{
	variable best_tag = @TagInfo_Type;
	variable best_score = -1;
	variable current_tag, current_score;
	variable tag_list;

	best_tag.file = "";
	best_tag.proto = "";
	best_tag.n = -1;

	tag_list = massoc_get(TagsArray, tag);
	if (tag_list != NULL) {
		foreach (tag_list) {
			current_tag = ();
			current_score = ntte_calculate_score(current_tag.data);
			if (current_score > best_score) {
				best_score = current_score;
				best_tag = current_tag.data;
			}
		}
	}

	return best_tag;
}

static define ntte_goto_tag(tag)
{
        bol();
        () = fsearch(tag);

        walk_goto_current_position();
}

% search this tag.
static define ntte_follow_tag(file, proto, tag, n)
{
        variable CTags_Save_Case_Search;

        if ((n < 0) or (strlen(tag) == 0))
                verror("ctags: Tag (%s) Not Found.", tag);

        !if (read_file(file))
                verror("ctags: Error opening file: %s.", file);

        if (n > 0) {
                goto_line(n);
        } else {
                CTags_Save_Case_Search = CASE_SEARCH;

                if (Ntte_Case_Sensitive >= 0)
                        CASE_SEARCH = Ntte_Case_Sensitive;

                if (Ntte_Look_From_End) {
                        eob();
                        bol_bsearch(proto);
                } else {
                        bob();
                        bol_fsearch(proto);
                }

                CASE_SEARCH = CTags_Save_Case_Search;

                !if (())
                        verror("ctags: Error looking for tag (%s)[%s].", tag, proto);
        }
        ntte_goto_tag(tag);

        return 0;
}

static define ntte_next_match()
{
}

static define ntte_store_pos(tag, file, pos)
{
	NtteOldTag = tag;
	NtteOldFile = file;
	NtteOldPos = pos;
}


define print_struct (s)
{
	variable name, value;

	foreach (get_struct_field_names (s))
	{
                name = ();
                value = get_struct_field (s, name);
                vinsert("s.%s = %s\n", name, string(value));
	}
}

static define ntte_find_tag(tag)
{
        variable start_file;
	variable ti = @TagInfo_Type;

        ERROR_BLOCK {
                walk_return_to_marked_position();
        }

        getbuf_info();
        pop; pop; exch;
        start_file = expand_filename(dircat());

        % Don't do it if we haven't moved.
        if (andelse  { NtteOldTag == tag }
	    { NtteOldPos == what_line() }
            { strcmp(start_file, NtteOldFile) == 0 }) {
                ntte_next_match();
                message("ntte: tag is Here!");
                return;
        }

        % Save Current Pos
        walk_mark_current_position();

        % Update Buffer
        ntte_update_array();

        ti = ntte_find(start_file, tag);

%	setbuf("*scratch*"); eob(); print_struct(ti);

        if ((ti.n < 0) and (Ntte_Never_Ask == 0)) {
                % Tag not found, ask user (if requested) and retry.
                tag = strtrim(read_mini("(Tag Not found) Find tag:", Null_String, tag));
                if (strlen(tag))
                        ti = ntte_find(start_file, tag);
        }

        ntte_follow_tag(ti.file, ti.proto, tag, ti.n);

        ntte_store_pos(tag, ti.file, what_line());

        % Now I can Store the old pos permanently.
        walk_store_marked_position();
}
%}}}

% I Don't like this name, but I retain it for compatibility.
% Find Tag Under cursor.
public define _find_tag()
{
        variable tag;

        % Get Tag
        tag = ntte_get_tag();

        !if (strlen(tag))
                return;

        if (Ntte_Case_Sensitive == 0)
		tag = strup(tag);

        ntte_find_tag(tag);
}
local_setkey("_find_tag", ".");
setkey("_find_tag", ".");

definekey("_find_tag", ".", "C");
