% Minesweeper - a procrastination aid for JED
% Written 17/11/98, obscurity@obscure.org

%_debug_info = 1;
% debugmode = "step";

% Alter the next 5 variables to taste . . .
variable MS_no_rows = 15;
variable MS_no_cols = 40;
variable MS_no_mines = 100;      % Note that if you don't have enough mines,
variable MS_search_limit_x = 5;  % the recursive-search routine (for when you
variable MS_search_limit_y = 10; % land on a space) may give a stack overflow.
                                 % Either increase the number of mines, or
                                 % decrease the search limits to compensate
                                 % for this.

% Define keymap.  Default keys are :
%                  <space> : Try this position.
%                  <enter> : Mark this position as containing a mine.
%                  ?       : Mark this position as possibly a mine
%                  u       : Unmark this position
%                  q       : Quit game.
%  Plus whatever your usual navigation keys are.
!if (keymap_p("MS_keymap"))
      {
                 make_keymap("MS_keymap");
                 definekey("MS_guess"," ","MS_keymap");
                 definekey("MS_mark_place(\"!\")","^M","MS_keymap");
                 definekey("MS_mark_place(\"?\")","?","MS_keymap");
                 definekey("MS_unmark","u","MS_keymap");
                 definekey("MS_quit","q","MS_keymap");
      }

% Set up the colours.  These work well for my personal colour scheme;
% You may want to change them to suit your tastes.
create_syntax_table ("MS");
define_syntax ("#", '0', "MS");
define_syntax ("0-9", 'w', "MS");
define_syntax ("!?",'+', "MS");

variable MS_minefield = String_Type [MS_no_rows];
variable MS_places_left;
variable MS_minefield_init_string = "";

%If anyone knows a better way of padding a string to a specific length,
%with a specific character, let me know...
loop (MS_no_cols)
{
   MS_minefield_init_string = MS_minefield_init_string + "#";
}

% init random seed. This hangs jed on my linux box :( so I've commented it out
() = random (-1, 0);

define max(x,y)
{
   if (x>y) { return x; }
   else { return y; }
}

define min(x,y)
{
   if (x<y) { return x; }
   else { return y; }
}

define MS_init_minefield()
{
   variable x,y,xx,yy;
   variable count;

        for (x = 0; x < MS_no_rows; x++)
        {
           MS_minefield[x] = MS_minefield_init_string;
        }

% Lay mines.
   loop(MS_no_mines)
         {
                x = random(0, MS_no_rows+1) - 1;
                y = random(0,1);   % For some bizarre reason, without this
                                   % superflous call to random(), y always
                                   % ended up as an even number. Go figure...
                y = random(0, MS_no_cols) + 1;

                MS_minefield[x] = strsub(MS_minefield[x],y, '!');        % lay mine
         }

% Count mines
        for (x = 0; x < MS_no_rows; x++)
        {
          for (y = 1; y < MS_no_cols + 1; y++)
                 {
                        if (substr(MS_minefield[x], y, 1) != "!")
                                 {
                                        count = 0;
                                        for (xx = max(x-1,0); xx <= min(x+1,MS_no_rows-1);xx++)
                                          {
                                                 for(yy = max(y-1,1); yy < min(y+2,MS_no_cols+1);yy++)
                                                   {
                                                          if (substr(MS_minefield[xx],yy,1) == "!")
                                                                {
                                                                   count++;
                                                                }
                                                   }
                                          }

                                        if (count > 0)
                                          {

                                                MS_minefield[x] = strsub(MS_minefield[x],y, int(string(count)));
                                          }
                                 }
                 }
        }
   MS_places_left = (MS_no_rows*MS_no_cols) - MS_no_mines;
}

% Draw the game area.
define MS_draw_area()
{
   variable count;

   pop2buf("*minesweeper*");
   set_readonly(0);
   onewindow;
   erase_buffer();

   for (count = 0; count < MS_no_rows; count++)
         {
                insert(MS_minefield_init_string + "\n");
         }

   use_syntax_table("MS");

% get rid of the last newline
   eob();
   call("previous_char_cmd");
   del();
%and go to middle of area
   bob();
   set_readonly(1);
   goto_line(MS_no_rows / 2);
   goto_column(MS_no_cols / 2);
}

variable MS_what_found, MS_orig_found;  % Global to cut down on local stack
variable MS_orig_x, MS_orig_y;

define MS_try_pos();          % define for recursion
define MS_try_pos(curr_line, curr_col)
{
   variable x,y;
   goto_line(curr_line+1);
   goto_column(curr_col);

   MS_what_found = substr(MS_minefield[curr_line],curr_col,1);
   MS_orig_found = MS_what_found;
   if (what_char() == '#')
         {
                MS_places_left--;
         }
%  A select/case test would be better for MS_orig_found, but the recursion
%  causes too many nested selects :(
   if (MS_orig_found == "!")
         {
                MS_places_left = -1; % Game over
                return;
         }
   if (MS_orig_found == "#")
         {
                MS_what_found = " ";
                MS_minefield[curr_line] = strsub(MS_minefield[curr_line],curr_col, ' ');

                () = replace_chars(1,MS_what_found);
%       We're on an empty place, so check adjacent places too (recursivly).
                for (x = max(curr_line-1,0); x <= min(curr_line+1,MS_no_rows-1);x++)
                  {
                         for(y = max(curr_col-1,1); y <= min(curr_col+1,MS_no_cols+1);y++)
                           {
                                  MS_what_found = substr(MS_minefield[x],y,1);
%                                 Limit search so we don't overflow the stack
                                  if ((x > (MS_orig_x - MS_search_limit_x)) and
                                          (x < (MS_orig_x + MS_search_limit_x)) and
                                          (y > (MS_orig_y - MS_search_limit_y)) and
                                          (y < (MS_orig_y + MS_search_limit_y)))
                                        {
                                           if (MS_what_found == "#")
                                                 {
                                                        MS_try_pos(x,y);
                                                 }
                                           else
                                                 {
                                                        goto_line(x+1);
                                                        goto_column(y);
                                                        if (y <= MS_no_cols)
                                                          {
                                                                 if (what_char() == '#')
                                                                   {
                                                                          MS_places_left--;
                                                                   }
                                                                 () = replace_chars(1,MS_what_found);
                                                          }
                                                 }
                                        }
                           }
                  }
         }
   if ((MS_orig_found != "!") and
           (MS_orig_found != "#"))
         {
            replace_chars(1,MS_orig_found);
                () = left;
         }
}

% They blew it, so show them what the minefield looks like.
define MS_show_minefield()
{
   variable count;

   set_readonly(0);
   erase_buffer();

   for (count = 0; count < MS_no_rows; count++)
         {
                insert(MS_minefield[count] + "\n");
         }
   bob();
   set_readonly(1);

}

define MS_mark_place(markstr)
{
   if ((what_char() != '#') and
           (what_char() != '!') and
           (what_char() != '?'))
         {
                message("Don't be silly - you already know what's there.");
         }
   else
         {
                set_readonly(0);
                replace_chars(1,markstr);
                () = left;
                set_readonly(1);
         }
}

define MS_unmark()
{
   if ((what_char() != '!') and
           (what_char() != '?'))
         {
                message("You must be hallucinating - there's no mark there to remove.");
         }
   else
         {
                set_readonly(0);
                replace_chars(1,"#");
                () = left;
                set_readonly(1);
         }
}

define MS_quit()
{
   pop2buf("*minesweeper*");  % Just to be on the safe side.
   ungetkey('k');  % Set next key stroke to 'k' for kill
   delbuf("*minesweeper*");
   message("Princess Di would be ashamed of you!   (If she weren't dead, that is.)");
}

define minesweeper()
{
   MS_init_minefield();
   MS_draw_area();
   set_status_line("Minesweeper---[Mines:" + string(MS_no_mines)
                                   + "]---[Spaces:" + string(MS_places_left)
                                   + "]--- %p,%c ---%t",0);

   use_keymap("MS_keymap");
}

define MS_again(promptstr)
{
   variable try_again = "";

   while ((try_again != "Y") and
                  (try_again != "y") and
                  (try_again != "N") and
                  (try_again != "n"))
         {
                try_again = read_mini(promptstr + "  Play again ? (Y/N)","Y",try_again);
         }

   if ((try_again == "Y") or
           (try_again == "y"))
         {
                if (MS_places_left == 0)
                  {
                         if (MS_no_mines < 10)
                           {
                                  MS_no_mines++;
                           }
                         else
                           {
                                  MS_no_mines += MS_no_mines/10;
                           }
                         message("More mines added.  Good luck!");
                  }
                else
                  {
                         message("This time, try to remember to *avoid* the mines!");
                  }
                minesweeper();
         }
   if ((try_again == "N") or
           (try_again == "n"))
         {
                MS_quit();
         }
}

define MS_guess()
{
   variable curr_line = what_line-1;
   variable curr_col = what_column;

   MS_orig_x = curr_line;
   MS_orig_y = curr_col;
   if (curr_col > MS_no_cols)      % They're on the newline character.
         {
                message ("You've left the minefield. Coward!");
                return;
         }
   set_readonly(0);
   MS_try_pos(curr_line, curr_col);
   set_readonly(1);

   set_status_line("Minesweeper---[Mines:" + string(MS_no_mines)
                                   + "]---[Spaces:" + string(MS_places_left)
                                   + "]--- %p,%c ---%t",0);


   goto_line(curr_line+1);
   goto_column(curr_col);
   if (MS_places_left == 0)
         {
                MS_again("Congratulations, you win!");
         }
   if (MS_places_left == -1)
         {
                MS_show_minefield();
                MS_again("Boom!  You're dead.");
         }
}
add_completion("minesweeper");
