%%
%%  Ispell interface
%%
%%  2001/07/01 Modified by Francesc Rocher
%%
%%  Change-Log:
%%        - Totally new way to select words.
%%        - Added custom variable 'ispell_language'.
%%        - Added custom variable 'ispell_top_line'.
%%        - Words can be viewed sorted alphabetically or as they have
%%          been suggested by {a,i}spell.
%%        - Accented vowels are also recognized, as well as some spanish
%%          and catalan consonants (ccedilla, ntilde). Updates are welcome.
%%        - White characters are skipped after {a,i}spelling words.
%%
_debug_info=1;
_traceback=1;

custom_variable( "ispell_language", "english" );
custom_variable( "ispell_top_line", 1 );

static variable
    ibuf = "*ispell*",
    icol = color_number( "menu_selection" ),
    letters = "a-zA-Z";

static variable buf, file, num_win, old_buf;
static variable word, cmd, p, num, n, new_word;
static variable imark;

public define ispell_quit()
{
    sw2buf( old_buf );
    skip_white();
    pop2buf( buf );
    if( num_win == 1 )
        onewindow();
    set_buffer_modified_flag( 0 );
    delbuf( ibuf );
}

public define ispell_insert_word()
{
    set_readonly(0);
    goto_column( 8 );
    push_mark_eol(); trim(); new_word = bufsubstr();
    sw2buf( old_buf );
    pop2buf( buf );
    bskip_chars( letters ); push_mark();
    skip_chars( letters ); del_region();
    insert( new_word );
    ispell_quit();
}

static define ispell_update_hook()
{
    if( ispell_top_line == 1 )
    {
        if( what_line() == 1 )
            goto_line( 2 );
    }
    goto_column( 6 );
    imark = create_line_mark( icol );
}

static define ispell_no_hook();

static define ispell_sort( how )
{
    imark = NULL;
    set_buffer_hook( "update_hook", &ispell_no_hook );
    set_readonly( 0 );
    goto_column( 1 );
    push_mark_eol();
    word = bufsubstr();
    if( ispell_top_line )
        goto_line( 2 );
    else
        goto_line( 1 );
    if( how == "alpha" )
        goto_column( 7 );
    else
        goto_column( 1 );
    push_mark();
    eob();
    sort();
    eob(); call( "backward_delete_char" );
    () = bsearch( word );
    set_readonly( 1 );
    set_buffer_modified_flag( 0 );
    set_buffer_hook( "update_hook", &ispell_update_hook );
}

public define ispell_sort_alpha()
{
    ispell_sort( "alpha" );
}

public define ispell_sort_suggest()
{
    ispell_sort( "suggest" );
}

define ispell()
{

#ifdef OS2
   file = make_tmp_file("jedt");
#else
   file = make_tmp_file("/tmp/jed_ispell");
#endif

   buf = whatbuf();

   skip_chars(letters); bskip_chars(letters); push_mark(); % push_mark();

   n = _get_point ();
   skip_chars(letters);
   if (_get_point () == n)
     {
	pop_mark_0 (); %pop_mark_0 ();
	return;
     }

   %word = bufsubstr();
#ifdef MSDOS MSWINDOWS WIN32
   () = system(sprintf("echo %s | ispell -d %s -a > %s",
                       bufsubstr(), ispell_language, file));
#else
   if (pipe_region(sprintf("ispell -d %s -a > %s", ispell_language, file)))
     error ("ispell process returned a non-zero exit status.");
#endif

   setbuf(ibuf);
   set_readonly(0);
   erase_buffer();
   () = insert_file(file);
   () = delete_file(file);

   %%
   %% parse output
   %%
   bob();
   if (looking_at_char('@'))   % ispell header
     {
	del_through_eol ();
     }

   if (looking_at_char('*') or looking_at_char('+'))
     {
	message ("Correct");   % '+' ==> is derived from
        set_buffer_modified_flag( 0 );
        delbuf( ibuf );
%        sw2buf( old_buf );
        skip_white();
	return;
     }

   if (looking_at_char('#'))
     {
        set_buffer_modified_flag( 0 );
        delbuf( ibuf );
%        sw2buf( old_buf );
        skip_white();
	return (message("No clue."));
     }

   del(); trim(); eol_trim(); bol();
   if (ffind_char (':'))
     {
	skip_chars(":\t ");
	push_mark();
	bol();
	del_region();
     }

   insert ("   1   ");
   n = 2;
   while (ffind_char (' '))
     {
	go_left_1 ();
	if (looking_at_char(',')) del(); else go_right_1 ();
	trim(); newline();
	vinsert (" %3d   ", n);
	++n;
     }

   eob();
   del_through_eol ();
   () = up(1);
   del_through_eol ();
   call("backward_delete_char");
   bob();
   if(ispell_top_line == 1)
        vinsert("ISPELL [%s] <Return>: select word;  a,s: sort;  q: quit\n",
                ispell_language);
   num_win = nwindows();
   pop2buf(buf);
   old_buf = pop2buf_whatbuf(ibuf);

   ERROR_BLOCK
     {
	sw2buf(old_buf);
	pop2buf(buf);
	if (num_win == 1) onewindow();
        set_buffer_modified_flag( 0 );
        delbuf(ibuf);
     }

   set_buffer_modified_flag(0);
   set_readonly(1);
   set_buffer_hook ("update_hook", &ispell_update_hook);
   set_column_colors (icol,1,6);
   $0 = "ispell";
   !if(keymap_p($0))
    {
        make_keymap($0);
        definekey ("ispell_insert_word", "\r", $0);
        definekey ("ispell_sort_alpha", "a", $0);
        definekey ("ispell_sort_suggest", "s", $0);
        definekey ("ispell_quit", "q", $0);
    }
   use_keymap($0);
}
