% This is a simple shell mode.  It does not defined any form of indentation
% style.  Rather, it simply implements a highlighting scheme.
%
% Fri Sep 7 19:33:56 CEST 2001, Dino Leonardo Sangoi: Added a simple indentation
% support.
%
_debug_info=1;
_traceback=1;

implements("SH");

static variable SHName = "SH";
custom_variable("SH_INDENT", 8);

static define has_match()
{
	variable ret = 1;
	ERROR_BLOCK {
		_clear_error();
		ret = 0;
	}
	call("goto_match");
	return ret;
}

% Check if this work on current line opens a new indentation block.
static define is_opening(word)
{
	variable c;

	if (word != "")
		return (orelse { word == "then" }
			{ word == "do" }
			{ word == "in" }
			{ word == "else" }
			{ word == "elif" }
			);
	EXIT_BLOCK {
		pop_spot();
	}
	push_spot();
	go_left_1();
	c = what_char();
	if ((c == '{') or (c == '('))
		return 1;
	if (c == ')') {
		if (has_match())
			return 0;
		return 1;
	}
	return 0;
}

% Check if this work on current line closes an indentation block.
% (if back is set, Point is after interesting characters).
static define is_closing(word, back)
{
	variable c;

	if (word != "")
		return (orelse { word == "fi" }
			{ word == "done" }
			{ word == "esac" }
			{ word == "else" }
			{ word == "elif" }
			);
	EXIT_BLOCK {
		pop_spot();
	}
	push_spot();
	if (back)
		go_left_1();
	c = what_char();
	if ((c == '}') or (c == ')'))
		return 1;
	if (back)
		go_left_1();
	if (looking_at(";;"))
		return 1;
	return 0;
}

% Goto start of comment
static define sh_bskip_comment ()
{
	while (parse_to_point() != 0) {
		if (bfind_char('#') == 0) {
			bol;
			break;
		}
	}
}

public define sh_indent_line ()
{
        variable level = 0;
	variable word1= "", word2;
	variable cont = 0, count = 0;
	variable ptp;

	push_spot();
	eol();
	if (parse_to_point() == -2) {
		bol;
		if (looking_at("#")) {
			pop_spot();
			return; % Don't touch comments at first column.
		} else {
			eol;
			sh_bskip_comment();
		}
	}
	% Don't try too hard to handle strings by now. Insead, do a simple
	% check to avoid indenting multiline strings.
	bol();
	if (parse_to_point() == -1) {
		pop_spot();
		return;
	}

	% Search backward for an interesting word (or character sequence,
	% shell scripts are a strange beast).
	while (up_1) {
		eol();
		% Avoid comments or strings.
		ptp = parse_to_point();
		if (ptp == -2)
			sh_bskip_comment();
		else if (ptp == -1)
			continue;

		% Handle continuation lines.
		% This fail on things like (Real code follows):
                % echo "$BackupPatterns" | while read i ; do
		%    ls -a "$1" | grep "$i" 2> /dev/null \
		%       | while read line ; do
		%          echo "$1/$line"
		%       done
		%    done # <-- Bad! Should be at "echo" level!
		%    IFS="$SaveIFS" # <-- and the error propagates.
		if (blooking_at("\\")) {
			if (count == 0) {
				go_up_1; eol;
				blooking_at("\\");
				go_down_1;
				if (() == 0)
					level += SH_INDENT;
				break;
			} else {
				go_down_1();
				level -= SH_INDENT;
				break;
			}
		}
		% Search back for meaningful words
		bskip_chars(" \t");
		word1 = get_word();
		if (is_opening(word1)) {
			level += SH_INDENT;
			break;
		}
		if (is_closing(word1, 1))
			break;
		count++;
	}

	bol_skip_white();
	level += what_column() - 1;

        pop_spot();
        bol_skip_white();

	word2 = get_word();
	if (is_closing(word2, 0)) {
		level -= SH_INDENT;
	}

        flush("level2: "+word1+","+word2+" - "+string(level));
        if (level < 0)
                level = 0;

        push_mark();
        bol();
        del_region();
        insert_spaces(level);
}

create_syntax_table (SHName);
define_syntax ("#", "", '%', SHName);
define_syntax ("([{", ")]}", '(', SHName);

% Unfortunately, the editor cannot currently correctly deal with multiple
% string characters.  So, inorder to handle something like:
%    echo "I'd rather be home"
% make the '"' character the actual string character but also give '\''
% a string syntax.  However, this will cause '"' to give problems but
% usually, '"' characters will be paired.
define_syntax ('\'', '"', SHName);
define_syntax ('"', '"', SHName);

define_syntax ('\\', '\\', SHName);
define_syntax ("-0-9a-zA-Z_", 'w', SHName);        % words
define_syntax ("-+0-9", '0', SHName);   % Numbers
define_syntax (",;:", ',', SHName);
define_syntax ("%-+/&*=<>|!~^", '+', SHName);

#ifdef HAS_DFA_SYNTAX
%%% DFA_CACHE_BEGIN %%%
static define setup_dfa_callback (name)
{
	dfa_enable_highlight_cache ("shmode.dfa", name);
	dfa_define_highlight_rule ("\\\\.", "normal", name);
	dfa_define_highlight_rule ("#.*$", "comment", name);
	dfa_define_highlight_rule ("\"([^\\\\\"]|\\\\.)*\"", "string", name);
	dfa_define_highlight_rule ("\"([^\\\\\"]|\\\\.)*$", "string", name);
	dfa_define_highlight_rule ("'[^']*'", "string", name);
	dfa_define_highlight_rule ("'[^']*$", "string", name);
	dfa_define_highlight_rule ("[\\|&;\\(\\)<>]", "Qdelimiter", name);
	dfa_define_highlight_rule ("[\\[\\]\\*\\?]", "Qoperator", name);
	dfa_define_highlight_rule ("[^ \t\"'\\\\\\|&;\\(\\)<>\\[\\]\\*\\?]+",
				   "Knormal", name);
	dfa_define_highlight_rule (".", "normal", name);
	dfa_build_highlight_table (name);
}
dfa_set_init_callback (&setup_dfa_callback, SHName);
%%% DFA_CACHE_END %%%
#endif

() = define_keywords (SHName, "cddofiifin", 2);
() = define_keywords (SHName, "forletpwdset", 3);
() = define_keywords (SHName, "casedoneechoelifelseesacevalexitifeqreadtestthentype", 4);
() = define_keywords (SHName, "aliasbreakendifendswifdefifneqlocalshiftumaskunsetuntilwhile", 5);
() = define_keywords (SHName, "exportifndefreturnsetenvsourceswitch", 6);
() = define_keywords (SHName, "breaksw", 7);
() = define_keywords (SHName, "continuefunction", 8);

!if (keymap_p(SHName)) {
	make_keymap(SHName);
	definekey("sh_indent_line", "^I", SHName);
}

public define sh_mode ()
{
	set_mode(SHName, 0);
	use_syntax_table (SHName);
	use_keymap(SHName);
	mode_set_mode_info (SHName, "fold_info", "#{{{\r#}}}\r\r");
	set_buffer_hook ("indent_hook", "sh_indent_line");
	run_mode_hooks("sh_mode_hook");
}
