% Some semi-automatic tm style documentation generation.
%
% By Dino Leonardo Sangoi.
%
% If you hame my cext.sl, uncomment the next line and
% exchange c_top_of_function with new_top_of_function.
%
%require("cext");
%
_traceback=1;
_debug_info=1;
require("comments");

%!%+
%\function{tm_make_doc}
%\synopsis{Create documentation comments for a function}
%\usage{tm_make_doc()}
%\description
%   When this function is called from inside a C or SLang function, it
%   creates right above the function itself a set of commented lines
%   formatted as expected by tm2txt. Some data will be generated
%   automagically from the function definition.
%\notes
%   Bind to a key.
%\seealso{tm_set_attr}
%!%-
define tm_make_doc()
{
        variable s, name;
        variable c = get_comment_info();
        variable cb, cm, ce;

        if (c == NULL)
                return;

        % Uhm, the word should be defined by mode, I guess (but currently is not)
        define_word("A-Za-z0-9_");
        cb = strtrim(c.cbeg);
        ce = strtrim(c.cend);
        cm = strtrim(c.cbeg);
        if (c.cend != "") {
                if (strlen(cm) > 1)
                        cm = " " + cm[[1:]];
        }

%        new_top_of_function();
        c_top_of_function();
        !if (re_bsearch("[]\\)\\}\\;\\>\\\"]"))
                break;
        if (what_char != ')')
                return;

        right(1); % leave on Stack
        push_mark();
        go_left(());  % get from stack
        call("goto_match");
        bskip_chars(" \t\n");
        push_mark();
        bskip_word_chars();
        name = bufsubstr();
        bskip_chars("^\n");
        s = bufsubstr();
        bol();
        % TODO: Check if already exists, and update.

        % Grrr, Wy C uses " *%+" while slang uses "%!%+" ???
        if (cb == "%")
                insert (cb + "!%+\n");
        else
                insert (cb + "%+\n");
        vinsert("%s\\function{%s}\n", cm, name);
        insert (cm + "\\synopsis{...}\n");
        vinsert("%s\\usage{%s}\n", cm, s);
        insert (cm + "\\description\n");
        insert (cm + "   \n");
        insert (cm + "\\notes\n");
        insert (cm + "   \n");
        insert (cm + "\\seealso{...}\n");
        if (cm == "%")
                insert (cm + "!%-\n");
        else
                insert (cm + "%-\n");
        if (ce != "")
                insert(" "+ce+"\n");
}

%!%+
%\function{tm_set_attr}
%\synopsis{Add attribute around a word in tm format}
%\usage{tm_set_attr(String_Type attr)}
%\description
%   This function adds '\\attr{' and '}' around the word at the point.
%   (attr is replaced with the actual parameter).
%\notes
%   Bind this to a key. For example, I bind "tm_set_attr(\\"var\\")" to "^Wv".
%\seealso{tm_make_doc}
%!%-
define tm_set_attr(attr)
{
        % As abobe...
        define_word("A-Za-z0-9_");
        bskip_word_chars();
        vinsert("\\%s{", attr);
        skip_word_chars();
        insert("}");
}

provide("tmtools");
